VERSION 1.0 CLASS
BEGIN
  MultiUse = -1  'True
  Persistable = 0  'NotPersistable
  DataBindingBehavior = 0  'vbNone
  DataSourceBehavior  = 0  'vbNone
  MTSTransactionMode  = 0  'NotAnMTSObject
END
Attribute VB_Name = "eval"
Attribute VB_GlobalNameSpace = False
Attribute VB_Creatable = True
Attribute VB_PredeclaredId = False
Attribute VB_Exposed = False

'CLASS MODULE TO EVAULATE AN EXPRESSION

'This version recognises the main program variables listed below.
'x, y
'a, b, c, d
'l, m, (NB all but 'x' are presently commented out),
'together with the constants,
'pi and e.

'This is based on the code of the Microsoft document Q86688,
'http://support.microsoft.com/default.aspx?scid=kb;en-us;Q86688
 
'Modifications,  Philip Bolt, 2005
'Last updated: 14/8/05

Option Explicit
Option Compare Text

Private e_text As String            'Error text.
Private e_input As String           'Expression input string.
Private e_tok As String             'Current token kind.
Private f_nme As String             'Function' name.

Private e_spelling As String        'Current token spelling.

Private e_error As Boolean          'Expression error flag.
Private hide_mess As Boolean        'Flag to hide error messages.
Private eval_exists As Boolean      'Flag that an evaluation object exists.

Private pi As Double                'pi
Private e As Double                 'e

'Procedure name:    Class_Initialize
'Run when an evaluation object is created.

Private Sub Class_Initialize()

    If Not eval_exists Then
        hide_mess = False   '*** SET 'TRUE' TO HIDE ERROR MESSAGES ***
        pi = 4 * Atn(1)     'Define pi.
        e = Exp(1)          'Define e.
        eval_exists = True
    End If
        
End Sub

'Procedure name:    Class_Terminate
'Run when an evaluation object is terminated.

Private Sub Class_Terminate()
 
End Sub

'The evaulation method.

'Evaluates a string containing an infix numeric expression.
'If successful, returns True and places the result in 'value'.
'This is the top-level function in the expression evaluator.

Public Function eval(ByVal s As String, value As Double) As Integer

    'Initialize.
    e_error = False
    e_input = s
    
    e_nxt

   'Evaluate.
    value = e_prs(1)    'Return the evaluation through the parameter
 
   'Check for unrecognized input.
    If e_tok <> "" And Not e_error Then
        e_text = "Syntax error"
        e_error = True
    End If
    
    If e_error And Not hide_mess Then   'If there is an error and they are not to be hidden,
        MsgBox e_text, vbOKOnly, "EVALUATION ERROR"
        value = 0       'return zero.
    End If

    eval = Not e_error  'Return the inverted error flag as the function output.
   
End Function

' e_prs
'   Parse an expression, allowing operators of a specified
'   precedence or higher. The lowest precedence is 1.
'   This function gets tokens with 'e_nxt' and recursively
'   applies operator precedence rules.

Private Function e_prs(p As Integer) As Double

    Dim n As Double    ' Return value.
    Dim fun As String  ' Function name.

    On Error GoTo err_trap  'Set error trap for bad data.
    
   ' Parse expression that begins with a token (precedence 12).
    Select Case True
        Case e_tok = "num"  ' number.
            n = Val(e_spelling)
            e_nxt
        'CHECK FOR VARIABLE NAMES
        Case e_spelling = "x"           'This section returns from 'e_nxt' with
            n = x                       'e_tok' set to "id", so the variables must be
            e_nxt                       'processed before the routine gets to that
        Case e_spelling = "y"           'token.
            'n = y                      'This could be extended for a larger number of
            e_nxt                       'specified variable names.
        Case e_spelling = "a"
            'n = a
            e_nxt
        Case e_spelling = "b"
            'n = b
            e_nxt
        Case e_spelling = "c"
            'n = c
            e_nxt
        Case e_spelling = "d"
            'n = d
            e_nxt
        Case e_spelling = "l"
            'n = l
            e_nxt
        Case e_spelling = "m"
            'n = m
            e_nxt
        Case e_spelling = "pi"      'Include 'pi'
            n = pi
            e_nxt
        Case e_spelling = "e"       'Include 'e'
            n = e
            e_nxt
        Case e_tok = "-"            'unary minus.
            e_nxt
            n = -e_prs(11)          'Operand precedence 11.
        Case e_tok = "not"          'logical NOT.
            e_nxt
            n = Not e_prs(6)        'Operand precedence 6.
        Case e_tok = "("            'parentheses.
            e_nxt
            n = e_prs(1)
            e_match (")")
        Case e_tok = "id"           'Function or undefined variable.
            If check_var(e_spelling) Then 'it's a function.
                fun = e_spelling
                f_nme = fun
                e_nxt
                e_match ("(")
                n = e_prs(1)
                e_match (")")
                n = e_function(fun, n)
            End If
        Case Else
            If Not e_error Then
                e_text = "Syntax error, token = '" + e_spelling + "'"
                e_error = True
            End If
   End Select

   ' Parse binary operators.
    Do While Not e_error
        Select Case True
            Case p <= 11 And e_tok = "^"
                e_nxt
                n = n ^ e_prs(12)
            Case p <= 10 And e_tok = "*"
                e_nxt
                n = n * e_prs(11)
            Case p <= 10 And e_tok = "/"
                e_nxt
                n = n / e_prs(11)
            Case p <= 9 And e_tok = "\"
                e_nxt
                n = n \ e_prs(10)
            Case p <= 8 And e_tok = "mod"
                e_nxt
                n = n Mod e_prs(9)
            Case p <= 7 And e_tok = "+"
                e_nxt
                n = n + e_prs(8)
            Case p <= 7 And e_tok = "-"
                e_nxt
                n = n - e_prs(8)
            Case p <= 6 And e_tok = "="
                e_nxt
                n = n = e_prs(7)
            Case p <= 6 And e_tok = "<"
                e_nxt
                n = n < e_prs(7)
            Case p <= 6 And e_tok = ">"
                e_nxt
                n = n > e_prs(7)
            Case p <= 6 And e_tok = "<>"
                e_nxt
                n = n <> e_prs(7)
            Case p <= 6 And e_tok = "<="
                e_nxt
                n = n <= e_prs(7)
            Case p <= 6 And e_tok = ">="
                e_nxt
                n = n >= e_prs(7)
            Case p <= 5 And e_tok = "and"
                e_nxt
                n = n And e_prs(6)
            Case p <= 4 And e_tok = "or"
                e_nxt
                n = n Or e_prs(5)
            Case p <= 3 And e_tok = "xor"
                e_nxt
                n = n Xor e_prs(4)
            Case p <= 2 And e_tok = "eqv"
                e_nxt
                n = n Eqv e_prs(3)
            Case p <= 1 And e_tok = "imp"
                e_nxt
                n = n Imp e_prs(2)
            Case Else
                Exit Do
        End Select
        If e_error Then
            Exit Do
        End If
    Loop

    On Error GoTo 0
    
    e_prs = n
   
Exit Function

err_trap:

    e_error = True                  'Flag for an error.
    
    If Err.Number = 11 Then
        e_text = "Division by zero."
    Else
        e_text = "Error #" & Str(Err.Number) & " was generated by " _
                & vbCrLf & Err.Source & vbCrLf & Err.Description
    End If
    
    Err.Clear
   
End Function

'Routine to check for a syntactically valid variable name.
'All program variable names are checked before this is used, so
'if it is a valid name, it must be undefined.

'If a function name is found, True is returned.
'Function name: check_var(nme)
'Parameter:     nme;    the name to be checked.

Private Function check_var(nme)

    Dim l_n, j, c_cd As Integer, c_v, b_v As Boolean
    
    c_v = False      'Set a default return value.
    Select Case nme
        Case "abs", "atn", "cos", "exp", "fix", "int", "log", "rnd", "sgn", "sin", _
             "sqr", "tan", "log10", "log2", "rad", "deg", "sec", "cosec", "cot", _
             "arcsin", "arccos", "arcsec", "arccosec", "arccot", "sinh", "cosh", _
             "tanh", "sech", "cosech", "cotanh", "arcsinh", "arccosh", "arctanh", _
             "arcsech", "arccosech", "arccotanh"
             c_v = True 'Set return to flag that this is a function.
    End Select
    
    If Not c_v Then 'it should be a variable.
        b_v = False 'Clear the 'Bad variable' flag.
        
        l_n = Len(nme)  'store the length of the name.
    
        c_cd = Asc(Left(nme, 1))    'Store the first character.
        If (c_cd = 95) Or (c_cd > 47 And c_cd < 58) Then 'this is a digit or underscore.
            b_v = True 'Set the flag as the first char. is wrong.
            e_error = True
            e_text = "Illegal variable name, '" + nme + "'"
        Else
            For j = 1 To l_n    'Loop through the name.
                c_cd = Asc(Mid(nme, j, 1))  'Store the character.
                If (c_cd > 47 And c_cd < 58) Or (c_cd > 96 And c_cd < 123) Or c_cd = 95 Then 'valid character, so
                'IT'S OK SO DO NOTHING
                Else    'it's invalid, so
                    b_v = True      'Set the flag,
                    e_error = True  'flag the error and
                    e_text = "Illegal variable name, '" + nme + "'"
                    j = l_n         'terminate the loop.
                End If
            Next
            If Not b_v Then 'it's a valid name but not defined.
                e_error = True  'Flag the error
                e_text = "Undefined variable name, '" + nme + "'"
            End If
        End If
    End If
 
    check_var = c_v

End Function

' e_function.
'   Evaluate a function. This is a helper function to simplify
'   e_prs.

Private Function e_function(fun As String, arg As Double) As Double
 
    Dim n As Double

    On Error GoTo err_trap  'Set error trap for bad data.
    
    f_nme = fun

    Select Case LCase(fun)
        Case "abs": n = Abs(arg)
        Case "atn": n = Atn(arg)
        Case "cos": n = Cos(arg)
        Case "exp": n = Exp(arg)
        Case "fix": n = Fix(arg)
        Case "int": n = Int(arg)
        Case "log": n = Log(arg)
        Case "rnd": n = Rnd(arg)
        Case "sgn": n = Sgn(arg)
        Case "sin": n = Sin(arg)
        Case "sqr": n = Sqr(arg)
        Case "tan": n = Tan(arg)
        '***** DERIVED FUNCTIONS *****
        Case "log10": n = Log(arg) / Log(10)
        Case "log2": n = Log(arg) / Log(2)
        '***** RADIANS & DEGREES *****
        Case "rad": n = arg * pi / 180
        Case "deg": n = arg * 180 / pi
        '***** INVERSE TRIG FUNCTIONS *****
        Case "sec": n = 1 / Cos(arg)
        Case "cosec": n = 1 / Sin(arg)
        Case "cot": n = 1 / Tan(arg)
        Case "arcsin": n = Atn(arg / Sqr(1 - arg * arg))
        Case "arccos": n = Atn(-arg / Sqr(1 - arg * arg)) + 2 * Atn(1)
        Case "arcsec": n = Atn(arg / Sqr(arg * arg - 1)) + Sgn(arg - 1) * (2 * Atn(1))
        Case "arccosec": n = Atn(arg / Sqr(arg * arg - 1)) + (Sgn(arg) - 1) * (2 * Atn(1))
        Case "arccot": n = Atn(arg) + 2 * Atn(1)
        '***** HYPERBOLIC FUNCTIONS *********
        Case "sinh": n = (Exp(arg) - Exp(-arg)) / 2
        Case "cosh": n = (Exp(arg) + Exp(-arg)) / 2
        Case "tanh": n = (Exp(arg) + Exp(-arg)) / 2
        Case "sech": n = 2 / (Exp(arg) + Exp(-arg))
        Case "cosech": n = 2 / (Exp(arg) - Exp(-arg))
        Case "cotanh": n = (Exp(arg) + Exp(-arg)) / (Exp(arg) - Exp(-arg))
        Case "arcsinh": n = Log(arg + Sqr(arg * arg + 1))
        Case "arccosh": n = Log(arg + Sqr(arg * arg - 1))
        Case "arctanh": n = Log((1 + arg) / (1 - arg)) / 2
        Case "arcsech": n = Log((Sqr(1 - arg * arg) + 1) / arg)
        Case "arccosech": n = Log((Sgn(arg) * Sqr(1 + arg * arg) + 1) / arg)
        Case "arccotanh": n = Log((arg + 1) / (arg - 1)) / 2
        '************************************
        Case Else
            If Not e_error Then
                e_text = "Undefined function '" + fun + "'"
                e_error = True
            End If
    End Select

    On Error GoTo 0
    
    e_function = n
   
Exit Function

err_trap:

    e_error = True                  'Flag for an error.
    
    If Err.Number = 11 Then
        e_text = "Division by zero in function '" + fun + "'"
    Else
        e_text = "Error #" & Str(Err.Number) & " was generated by " _
                & vbCrLf & Err.Source & vbCrLf & Err.Description
    End If
    
    Err.Clear
   
End Function

' e_nxt
'   Get the next token into e_tok and e_spelling and remove the
'   token from e_input.
'   This function groups the input into "words" like numbers,
'   operators and function names.
'NB When  variables are included, they will exit this
'   procedure with the label in 'e-spelling' and 'e_tok' = "id".

Private Sub e_nxt()

    Dim is_keyword As Integer
    Dim c As String             ' Current input character.
    Dim is_id As Boolean
    
    If Not e_error Then
        e_tok = ""
        e_spelling = ""

        'Skip whitespace.
        Do
            c = Left(e_input, 1)
            e_input = Mid(e_input, 2)
        Loop While c = " " Or c = Chr(9) Or c = Chr(13) Or c = Chr(10)

        Select Case LCase(c)
            ' Number constant. Modify this to support hexadecimal, etc.
            Case "0" To "9", "."
                e_tok = "num"
                Do
                    e_spelling = e_spelling + c
                    c = Left(e_input, 1)
                    e_input = Mid(e_input, 2)
                Loop While (c >= "0" And c <= "9") Or c = "."
                e_input = c + e_input

            ' Variable name or keyword.
            Case "a" To "z", "_"
                e_tok = "id"
                Do
                    e_spelling = e_spelling + c
                    c = LCase(Left(e_input, 1))
                    e_input = Mid(e_input, 2)
                    is_id = (c >= "a" And c <= "z")
                    is_id = is_id Or c = "_" Or (c >= "0" And c <= "9")
                Loop While is_id
                e_input = c + e_input

                ' Check for keyword.
                is_keyword = -1
                Select Case LCase(e_spelling)
                    Case "and"
                    Case "eqv"
                    Case "imp"
                    Case "mod"
                    Case "not"
                    Case "or"
                    Case "xor"
                    Case Else: is_keyword = 0
                End Select
                If is_keyword Then
                    e_tok = LCase(e_spelling)
                End If

            ' Check for <=, >=, <>.
            Case "<", ">"
                e_tok = c
                c = Left(e_input, 1)
                If c = "=" Or c = ">" Then
                    e_tok = e_tok + c
                    e_input = Mid(e_input, 2)
                End If

            ' Single character token.
            Case Else
                e_tok = c
        End Select

        If e_spelling = "" Then
            e_spelling = e_tok
        End If
    End If 'not error.

End Sub

' e_match
'   Check the current token and skip past it.
'   This procedure helps with syntax checking.

Private Sub e_match(token As String)
 
    If Not e_error And e_tok <> token Then
        If token = "(" Then 'this is the start of a pair of parentheses.
            Select Case e_spelling
                'If the text is followed by an operator or space, it's a variable name.
                Case "", "+", "-", "*", "/", "\", "^", "mod", ">", "<", "=", _
                    "<=", ">=", "<>", "and", "or", "xor", "eqv", "imp"
                    e_text = "Undefined variable name, '" + f_nme + "'."
                Case Else   'there should be a left parenthesis.
                    If e_spelling = "" Then 'there's no character, so
                        e_text = "Missing '" + token + "'"
                    Else                    'It's the wrong character, so
                        e_text = "Expected " + token + ", got '" + e_spelling + "'"
                    End If
            End Select
        Else    'It's the end of a pair of parentheses.
            If e_spelling = "" Then 'there's no character, so
                e_text = "Missing '" + token + "'"
            Else                    'It's the wrong character, so
                e_text = "Expected " + token + ", got '" + e_spelling + "'"
            End If
        End If
        e_error = True
    End If
   
    e_nxt
   
End Sub


